/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group

******************************************************************/


#include <sys/types.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <dlfcn.h>
#include <locale.h>
#include "layout.h"
#include "layout_int.h"

#define DEFAULTPATH "/usr/lib/locale"
#define _POSIX_PATH_MAX 50

/*
 * NAME: nextpath
 *
 * FUNCTION:
 *	Get the next path element from the given LOCPATH.
 *	The result followed by '/' is stored in the buffer,
 *	and LOCPATH pointer is advanced.
 *
 * RETURN VALUE DESCRIPTION:
 *	Length of the next path element.
 */

int next_path(char **LOCPATH, char *buf)
{
	char	*locptr;
	int	len;


	locptr = *LOCPATH;
	while (*locptr && *locptr != ':')
            locptr++;
	len = locptr - *LOCPATH;
	if (len > _POSIX_PATH_MAX - 1)
		return -1;
	if (len == 0) {
		locptr++;
		buf[0] = '.';
		len = 1;
	}
	else {
		if (*locptr && !*++locptr)
			locptr--;
		(void)memcpy(buf, *LOCPATH, len);
	}
	*LOCPATH = locptr;
	return len;
}

/*
 * NAME: m_create_layout
 *
 * FUNCTION:
 *       This function is responsible to load the correct shared object
 *       according to the input locale, if success it'll return the 
 *       layout object.
 *
 * RETURN VALUE DESCRIPTION:
 *	Layout Object
 */

LayoutObject m_create_layout(locale, modifier)
const AttrObject locale;
const char *modifier;
{
    void        *handle    			= (void *)NULL;
    LayoutObj   (*initializeFunction)() 	= (LayoutObj(*)())NULL;
    LayoutObj   layoutObj                	= (LayoutObj)NULL;
    int         nameLen				= 0;
    int         methodNameLen  			= 0;
    int         defaultPath			= 0;
    char        *LOCPATH			= NULL;
    char        sharedPath[FILENAME_MAX];
    char        methodName[FILENAME_MAX];
    char        locale_name[FILENAME_MAX];
    char        *locpath_env                     = NULL;

    if (!locale)
        strcpy(locale_name, setlocale(LC_CTYPE, (char *) NULL));
    else
        strcpy(locale_name, locale);

    sprintf(methodName, "/%s.layout.so", locale_name);
    methodNameLen = strlen(methodName);
    
    LOCPATH = (char*) calloc(FILENAME_MAX, sizeof(char));
    locpath_env = getenv("LOCPATH");
    sprintf(LOCPATH, "%s", locpath_env);
    if (!locpath_env) { 
        sprintf(LOCPATH, "%s/%s", DEFAULTPATH, locale_name);
        defaultPath = TRUE;
    }
    else if (strcmp (LOCPATH, DEFAULTPATH) == 0)
        defaultPath = TRUE;

    while (TRUE) {
        if (*LOCPATH == '\0') {
            if (!defaultPath) {
                sprintf(LOCPATH, "%s/%s", DEFAULTPATH, locale_name);
                defaultPath = TRUE;
            } else {
                errno = EINVAL;
                break;
            }
        }
        nameLen = (int) next_path(&LOCPATH, &sharedPath[0]);
        if ((nameLen + methodNameLen) > FILENAME_MAX ) {
            errno = ENAMETOOLONG;
            break;
        }

        /* Check if shared library is available. */
        strcpy (&sharedPath[nameLen], methodName);

        if((handle = dlopen (sharedPath , RTLD_NOW))){
            initializeFunction = (LayoutObj(*)())dlsym(handle,"_LayoutObjectInit");
            break;
        }
    }
    /*initializeFunction = NULL;i*/
    if(!initializeFunction) {
        errno      = EINVAL;
        layoutObj  = (LayoutObject)DefaultOpen();
    }
    else
        layoutObj = (*initializeFunction)(locale_name);

    if(!layoutObj)
        return ((LayoutObj)NULL);
    else
        return(layoutObj);
}

/*
 * NAME: m_destroy_layout
 *
 * FUNCTION:
 *       This function is responsible to free the memory allocated 
 *       and used by the layout object.
 *
 * RETURN VALUE DESCRIPTION:
 *	 integer values either success or fail
 */

int
m_destroy_layout(layout_object)
    const LayoutObject  layout_object;
{
    LayoutObj           layout_obj = (LayoutObj) layout_object;
    int                 result     = -1;
  
    if (layout_obj) {
        result = layout_obj->methods->destroy(layout_obj);
    } else {
        errno = EFAULT;
    }
    
    return(result);
}

/*
 * NAME: m_transform_layout
 *
 * FUNCTION:
 *       This function is responsible of the transformation of 
 *       single byte text
 *       Reordering & Shaping
 *
 * RETURN VALUE DESCRIPTION:
 *	 integer values either success or fail
 */

int
m_transform_layout(layout_object, InpBuf, InpSize, OutBuf, OutSize,
                   InpToOut, OutToInp, Property, InpBufIndex)
    LayoutObject                layout_object;
    const char                  *InpBuf;
    size_t                      InpSize;
    void                        *OutBuf;
    size_t                      *OutSize;
    size_t                      *InpToOut;
    size_t                      *OutToInp;
    unsigned char               *Property;
    size_t                      *InpBufIndex;

{  
    LayoutObj                   layout_obj = (LayoutObj) layout_object;
    int                         result     = -1;
  
    if (layout_obj) {
        result = layout_obj->methods->transform(layout_obj, InpBuf, InpSize,
                                                OutBuf, OutSize, InpToOut,
                                                OutToInp, Property,InpBufIndex);
    } else
        errno = 0;

    return(result);

}

/*
 * NAME: m_wtransform_layout
 *
 * FUNCTION:
 *       This function is responsible of the transformation of 
 *       Unicode text
 *       Reordering & Shaping
 *
 * RETURN VALUE DESCRIPTION:
 *	 integer values either success or fail
 */

int
m_wtransform_layout(layout_object, InpBuf, InpSize, OutBuf, Outsize,
                   InpToOut, OutToInp, Property, InpBufIndex)
    LayoutObject                layout_object;
    const wchar_t               *InpBuf;
    const size_t                InpSize;
    void                        *OutBuf;
    size_t                      *Outsize;
    size_t                      *InpToOut;
    size_t                      *OutToInp;
    unsigned char               *Property;
    size_t                      *InpBufIndex;
{
    LayoutObj                   layout_obj = (LayoutObj) layout_object;
    int                         result     = -1;
   
    if (layout_obj) {
        result = layout_obj->methods->wcstransform(layout_obj, InpBuf,
                                                   InpSize, OutBuf, Outsize,
                                                   InpToOut, OutToInp,
                                                   Property, InpBufIndex);
    } else
        errno = EBADF;

    return(result);
}

/*
 * NAME: m_setvalues_layout
 *
 * FUNCTION:
 *       This function is responsible to set the attributes used 
 *       by the engine in the process of Reordering & Shaping.
 *
 * RETURN VALUE DESCRIPTION:
 *	 integer values either success or fail
 */

int
m_setvalues_layout(layout_object, values, index_returned)
    LayoutObject        layout_object;
    LayoutValues        values;
    int                 *index_returned;
{
    LayoutObj           layout_obj = (LayoutObj) layout_object;
    int                 result     = -1;

    if (layout_obj) {
        result = layout_obj->methods->setvalues(layout_obj, values, index_returned);
    } else {
        errno = EINVAL;
    }

    return(result);
} 

/*
 * NAME: m_getvalues_layout
 *
 * FUNCTION:
 *       This function is responsible to get the attributes used 
 *       by the engine in the process of Reordering & Shaping.
 *
 * RETURN VALUE DESCRIPTION:
 *	 integer values either success or fail
 */
int
m_getvalues_layout(layout_object, values, index_returned)
    LayoutObject        layout_object;
    LayoutValues        values;
    int                 *index_returned;
{
    LayoutObj           layout_obj = (LayoutObj) layout_object;
    int                 result     = -1;

    if (layout_obj) {
        result = layout_obj->methods->getvalues(layout_obj, values, index_returned);
    } else {
        errno = EINVAL;
        *index_returned = -1;
    }

    return(result);
} 
