/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <malloc.h>

#ifndef _THAI_TYPE_H_
#define _THAI_TYPE_H_

#define MAX_TRANSFORMED_CHAR    2

#define THAI_UC_START           0x0E00
#define THAI_UC_END             0x0E7F

#define IS_THAI_UC(wch)         (wch >= THAI_UC_START && wch <= THAI_UC_END)

#define THAI_NUM_0              0x0E50
#define THAI_NUM_9              0x0E59

#define IS_THAI_NUM(wch)        (wch >= THAI_NUM_0 && wch <= THAI_NUM_9)

#define ARABIC_NUM_0            0x0030
#define ARABIC_NUM_9            0x0039

#define IS_ARABIC_NUM(wch)      (wch >= ARABIC_NUM_0 && wch <= ARABIC_NUM_9)

/* Numeral Shaping Conversion */

#define THAI_NUM_SHAPE          1
#define ARABIC_NUM_SHAPE        (1 << 1)

#define ARABIC_THAI_DIFF        0x0E20

/* ----- Category of Thai Characters ----- */

typedef unsigned long   th_char_t;

/*
    Consonants
        * C     - normal Consonant
        * CA    - Consonant with Ascender
        * CRA   - Consonant with Right-positioned Ascender
        * CD    - Consonant with Descender
        * CUD   - Consonant with Unconnected Descender
*/

#define C               (1L)
#define CA              (1L << 1)
#define CRA             (1L << 2)
#define CD              (1L << 3)
#define CUD             (1L << 4)

/*
    Non-Base-Line Symbols (Vowels, Tone Marks, and Diacritics)
        * UV1   - upper-vowel-line type-1 Upper Vowel
        * UV2   - upper-vowel-line type-2 Upper Vowel
        * UV3   - upper-vowel-line type-3 Upper Vowel
        * LV1   - lower-vowel-line type-1 Lower Vowel
        * LV2   - lower-vowel-line type-2 Lower Vowel
        * TM    - tone-mark-line Tone Mark
        * UD1   - tone-mark-line type-1 Upper Diacritic
        * UD2   - tone-mark-line type-2 Upper Diacritic
        * UD3   - upper-vowel-line type-3 Upper Diacritic
        * LD    - lower-vowel-line Lower Diacritic
*/

#define UV1             (1L << 5)
#define UV2             (1L << 6)
#define UV3             (1L << 7)
#define LV1             (1L << 8)
#define LV2             (1L << 9)
#define TM              (1L << 10)
#define UD1             (1L << 11)
#define UD2             (1L << 12)
#define UD3             (1L << 13)
#define LD              (1L << 14)

/*
    Base-line Symbols (Vowels and other Thai symbols)
        * VLA   - Vowel with Left-hanging Ascender
        * VUA   - Vowel with Unconnected Ascender
        * VRD   - Vowel with Right Descender
        * NUM   - NUMeral
        * BS    - other Base-line Symbol
*/

#define VLA             (1L << 15)
#define VUA             (1L << 16)
#define VRD             (1L << 17)
#define NUM             (1L << 18)
#define BS              (1L << 19)

/* 
    Auxilirary Type
        * CONSONANT     - CONSONANT
        * CNA           - Consonant with No Ascender
        * CND           - Consonant with No Descender
        * UV            - Upper Vowel
        * LV            - Lower Vowel
        * NBV           - Non-Base-line Vowel
        * COMPOSING     - COMPOSING character
*/

#define CONSONANT       (C | CA | CRA | CD | CUD)
#define CNA             (C | CD | CUD)
#define CND             (C | CA | CRA)
#define UV              (UV1 | UV2 | UV3)
#define LV              (LV1 | LV2)
#define NBV             (UV | LV)
#define COMPOSING       (NBV | TM | UD1 | UD2 | UD3 | LD)

/* For glyps only (Not Applicable) */

#define NA              (1L << (sizeof(long)*8 - 1))

/* An offset to alternate glyps  */

/*
     Composing
        * CNA + TM              - glyph for upper-vowel-line TM/UD1/UD2
        * CRA + UV + TM         - glyph for left-shifted TM/UD1/UD2
        * CRA + TM              - glyph for left-shifted upper-vowel-line
                                  TM/UD1/UD2
        * CRA + UV              - glyph for left-shifted UV/UD3
        * CD  + LV              - glyph for lower-position LV/LD
        * CA  + UV              - glyph for smaller-size CA
        * CUD + LV              - glyph for descender-absent CUD
*/

#define CNA_TM_OFFSET           0x0021
#define CRA_UV_TM_OFFSET        0x0028
#define CRA_TM_OFFSET           0x002F
#define CRA_UV_OFFSET           0x0030
#define CA_UV_OFFSET            0x0051
#define CUD_LV_OFFSET           0x0053
#define CD_LV_OFFSET            0x0004

/*
    Kerning
    * NOASCENDER + VLA      - glyph for left-shifted VLA
*/

#define NOASC_VLA_OFFSET        0x001B

/* Type Table corresponding to UCS-2 Charset (0x0E00 - 0x0E7F) */

static th_char_t ThaiType_tbl[] = {
        NA,     C,      C,      C,      C,      C,      C,      C,
        C,      C,      C,      C,      C,      CUD,    CD,     CD,
        CUD,    C,      C,      C,      C,      C,      C,      C,
        C,      C,      C,      CRA,    C,      CRA,    C,      CRA,
        C,      C,      C,      C,      VRD,    C,      VRD,    C,
        C,      C,      C,      C,      CA,     C,      CA,     BS,
        BS,     UV2,    BS,     VUA,    UV1,    UV3,    UV2,    UV3,
        LV1,    LV2,    LD,     NA,     NA,     NA,     NA,     BS,
        BS,     BS,     VLA,    VLA,    VLA,    BS,     BS,     UD2,
        TM,     TM,     TM,     TM,     UD1,    UD1,    UD3,    BS,
        NUM,    NUM,    NUM,    NUM,    NUM,    NUM,    NUM,    NUM,
        NUM,    NUM,    BS,     BS,     NA,     NA,     NA,     NA,
        NA,     NA,     NA,     NA,     NA,     NA,     NA,     NA,
        NA,     NA,     NA,     NA,     NA,     NA,     NA,     NA,
        NA,     NA,     NA,     NA,     NA,     NA,     NA,     NA,
        NA,     NA,     NA,     NA,     NA,     NA,     NA,     NA
};

#define TYPE(wch)               (ThaiType_tbl[wch - THAI_UC_START])

#define W_SPACE                 0x0020
#define NIKKHAHIT               0x0E4D
#define SARA_AA                 0x0E32
#define LAAK_KANG               0x0E45

#define NEW_CELL                (1 << 7)

#endif
